unit UIDFC;

//    This file is part of IDFC - the i.LON Data File Converter.
//
//    Copyright (C) 2010-2017  TEROC AB.
//
//    IDFC is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    IDFC is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with IDFC.  If not, see <http://www.gnu.org/licenses/>.
//
//    For more information, contact Sven Ruin, sven.ruin@teroc.se

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Buttons, StdCtrls, ExtCtrls, Grids;

type
  TForm1 = class(TForm)
    OpenDialog1: TOpenDialog;
    Memo1: TMemo;
    Label1: TLabel;
    Timer1: TTimer;
    StringGrid1: TStringGrid;
    Button1: TButton;
    Button2: TButton;
    procedure Timer1Timer(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure StringGrid1SelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
  private
    { Private declarations }
	procedure Init;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

uses Unit2;

{$R *.DFM}

type
  TLogRec = record
    Fi1,Fi2,Fi3,Fi4,Fi5,Fi6,Fi7: string;
    NFields: Integer;
  end;

var
  NJobs: Integer = 0;
  Typ: array[1..9] of Integer;
  InFileName, OutFileName: array[1..9] of string;
  Interv, NCols, PrevAge: array[1..9] of Integer;
  ColName, Uni: array[1..40,1..9] of string;

procedure TForm1.Init;
var
  InitFile: TextFile;
  i,j: Integer;
  Tmp: string;
begin
  StringGrid1.Cells[0,0] := 'Job number';
  StringGrid1.Cells[1,0] := 'Type';
  StringGrid1.Cells[2,0] := 'Output file';
  StringGrid1.Cells[3,0] := 'Time interval for output file [s]';
  StringGrid1.Cells[4,0] := 'Number of columns in output file';
  StringGrid1.Cells[5,0] := 'Latest data';
  if OpenDialog1.Execute then
  begin

    try

      AssignFile(InitFile, OpenDialog1.FileName);
      Reset(InitFile);
      Readln(InitFile,Tmp);
      if Copy(Tmp, 1, 5) = '[0,1]'
        then Readln(InitFile, NJobs)
        else Memo1.Lines.Add('Error finding initialization file section [0,1]');
      if NJobs > 9 then
      begin
        NJobs := 9;
        Memo1.Lines.Add('Number of jobs limited to 9');
      end;
      StringGrid1.RowCount := NJobs + 1;
      for i := 1 to NJobs do
      begin
        StringGrid1.Cells[0,i] := IntToStr(i);
        Readln(InitFile,Tmp);
        if Copy(Tmp, 1, 5) = '[' + IntToStr(i) + ',1]' then
        begin
          Readln(InitFile,Typ[i]);
        end
        else
        begin
          Typ[i] := 0;
          Memo1.Lines.Add('Error finding initialization file section [' + IntToStr(i) + ',1]');
        end;
        case Typ[i] of
          0: StringGrid1.Cells[1,i] := 'Inactive';
          1: StringGrid1.Cells[1,i] := 'Timer conv.';
          2: StringGrid1.Cells[1,i] := 'Timer ana. & conv.';
          3: StringGrid1.Cells[1,i] := 'Manual';
        end; // case
        Readln(InitFile,Tmp);
        if Copy(Tmp, 1, 5) = '[' + IntToStr(i) + ',2]'
          then Readln(InitFile, InFileName[i])
          else Memo1.Lines.Add('Error finding initialization file section [' + IntToStr(i) + ',2]');
        Readln(InitFile,Tmp);
        if Copy(Tmp, 1, 5) = '[' + IntToStr(i) + ',3]'
          then Readln(InitFile, OutFileName[i])
          else Memo1.Lines.Add('Error finding initialization file section [' + IntToStr(i) + ',3]');
        StringGrid1.Cells[2,i] := OutFileName[i];
        Readln(InitFile,Tmp);
        if Copy(Tmp, 1, 5) = '[' + IntToStr(i) + ',4]'
          then Readln(InitFile, Interv[i])
          else Memo1.Lines.Add('Error finding initialization file section [' + IntToStr(i) + ',4]');
        StringGrid1.Cells[3,i] := IntToStr(Interv[i]);
        Readln(InitFile,Tmp);
        if Copy(Tmp, 1, 5) = '[' + IntToStr(i) + ',8]'
          then Readln(InitFile, NCols[i])
          else Memo1.Lines.Add('Error finding initialization file section [' + IntToStr(i) + ',8]');
        if NCols[i] > 0
          then StringGrid1.Cells[4,i] := IntToStr(NCols[i])
          else StringGrid1.Cells[4,i] := '?';
        Readln(InitFile,Tmp);
        if Copy(Tmp, 1, 5) = '[' + IntToStr(i) + ',9]' then
        begin
          for j := 1 to NCols[i] do
          begin
            Readln(InitFile, ColName[j,i]);
            Readln(InitFile, Uni[j,i]);
          end;
        end
        else Memo1.Lines.Add('Error finding initialization file section [' + IntToStr(i) + ',9]');
      end;
      Memo1.Lines.Add('Initialization file read');

    finally

      CloseFile(InitFile);

    end; // try-finally
  end; // if
  if Typ[StringGrid1.Row]=3 then Button2.Enabled := True else Button2.Enabled := False;
end; // Init

function TakeApart(InStr: string): TLogRec;
var
  i,j,k,L: Integer;
  Field: array[1..9] of string;
begin
  L := Length(InStr);
  k := 1;
  while (k < L) and (InStr[k] <> ',') do Inc(k);
  Field[1] := Copy(InStr,1,k-1);
  i := 1;
  while (k < L) and (i < 9) do
  begin
    Inc(i);
    if i = 2 then j := k + 2 else j := k + 3;
    k := j;
    while (k < L) and (InStr[k] <> #34) do Inc(k);
    Field[i] := Copy(InStr,j,k-j);
  end;
  TakeApart.Fi1 := Field[1];
  TakeApart.Fi2 := Field[2];
  TakeApart.Fi3 := Field[3];
  TakeApart.Fi4 := Field[5];
  TakeApart.Fi5 := Field[7];
  TakeApart.Fi6 := Field[8];
  TakeApart.Fi7 := Field[9];
  TakeApart.NFields := i;
end {TakeApart};

function RStr(InStr: string): string;
var
  L: Integer;
begin
  L := Length(InStr);
  if L < 28 then RStr := InStr else RStr := Copy(InStr, L-26, 27);
end;

function ChangeToPoint(const NumberString: string): string;
var
  ChangeString: string;
  i: Integer;
begin
  ChangeString := NumberString;
  for i:= 1 to Length(ChangeString) do
  begin
    if ChangeString[i] = DecimalSeparator then ChangeString[i] := '.';
  end;
  ChangeToPoint := ChangeString;
end;

function Analy(Job: Integer): Boolean; // ANALYZE INPUT FILE
var
  InFile: TextFile;
  InString,Identif: string;
  n,Li: Integer;
  InData: TLogRec;
begin

  try

    AssignFile(InFile,InFileName[Job]);
    Reset(InFile);
    Form1.Memo1.Lines.Add(TimeToStr(Now)+' Starting analysis');
    Form1.Memo1.Lines.Add('Job '+IntToStr(Job));
    Form1.Memo1.Lines.Add('Opening file: '+InFileName[Job]);
    Form1.Memo1.Lines.Add('New columns:');
    Li := 0;
    if NCols[Job] = 0 then
    begin
      ColName[1,Job] := 'TIMESTAMP';
      Uni[1,Job] := 'TS';
      ColName[2,Job] := 'RECORD';
      Uni[2,Job] := 'RN';
      NCols[Job] := 2;
    end;

    // Skip first line
    Readln(InFile,InString);

    // Identify data points
    while (not Eof(InFile)) do
    begin
      Readln(InFile,InString);
      InData := TakeApart(InString);
      if InData.NFields = 9 then
      begin
        n := 3;
        Identif := InData.Fi2;
        if Identif <> '' then
        begin
          while (Identif <> ColName[n,Job]) and (n <= NCols[Job]) do Inc(n);
          if n > NCols[Job] then
          begin
            Inc(NCols[Job]);
            ColName[NCols[Job],Job] := Identif;
            Uni[NCols[Job],Job] := InData.Fi6;
            Form1.Memo1.Lines.Add(IntToStr(NCols[Job])+' '+ColName[NCols[Job],Job]);
          end;
        end;
      end
      else
      begin
        Li := Li + 1;
      end;
    end;
    Form1.Memo1.Lines.Add('Number of input lines not accepted: '+IntToStr(Li));
    CloseFile(InFile);
    Form1.Memo1.Lines.Add('Succeeded, input file closed');
    Analy := True;

  except

    CloseFile(InFile);
    Form1.Memo1.Lines.Add('Failed, input file closed');
    Analy := False;

  end; // try-except
end; // Analy

function Conver(Job: Integer): string; // CONVERT AND WRITE TO OUTPUT FILE
var
  InFile,OutFile: TextFile;
  Col: array[1..40] of string;
  InString,OutString,TimeString0,TimeString1,Identif: string;
  n,Le,TimeStringLe1,TimeStringLe2,RN,Pass: Integer;
  InData: TLogRec;
  Wri: Boolean;
begin
  RN := 0;
  case Interv[Job] of
    1: begin TimeStringLe1 := 10; TimeStringLe2 :=  8; end;
    10: begin TimeStringLe1 :=  8; TimeStringLe2 :=  8; end;
    60: begin TimeStringLe1 :=  7; TimeStringLe2 :=  5; end;
    600: begin TimeStringLe1 :=  5; TimeStringLe2 :=  5; end;
    3600: begin TimeStringLe1 :=  4; TimeStringLe2 :=  2; end;
    86400: begin TimeStringLe1 :=  1; TimeStringLe2 :=  0; end;
  end; // case

  try

    AssignFile(InFile,InFileName[Job]);
    AssignFile(OutFile,OutFileName[Job]);
    Reset(InFile);
    Rewrite(OutFile);
    Form1.Memo1.Lines.Add(TimeToStr(Now)+' Starting conversion');
    Form1.Memo1.Lines.Add('Job '+IntToStr(Job));
    Form1.Memo1.Lines.Add('Writing to file: '+OutFileName[Job]);

    // Skip first line
    Readln(InFile,InString);

    // Write header
    Writeln(OutFile,'"TOA5","Station","Datalogger","","","","","Table"');
    OutString := '';
    for n := 1 to NCols[Job] do
    begin
      OutString := OutString + '"' + RStr(ColName[n,Job]) + '"';
      if (n < NCols[Job]) then OutString := OutString + ',';
    end;
    Writeln(OutFile,OutString);
    OutString := '';
    for n := 1 to NCols[Job] do
    begin
      OutString := OutString + '"' + Uni[n,Job] +'"';
      if (n < NCols[Job]) then OutString := OutString + ',';
    end;
    Writeln(OutFile,OutString);
    OutString := '';
    for n := 1 to NCols[Job] do
    begin
      OutString := OutString + '""';
      if (n < NCols[Job]) then OutString := OutString + ',';
    end;
    Writeln(OutFile,OutString);

    for n := 1 to NCols[Job] do Col[n] := '"NAN"';
    InString := '';
    TimeString1 := '';
    Pass := 1;
    Wri := False;

    // Write one line for each time step
    while (InString <> '000000000000000000000000') do
    begin
      if (not Eof(InFile))
        then Readln(InFile,InString)
        else
        begin
         if Pass = 2
          then InString := '000000000000000000000000' // End reading, writing
          else
          begin
            if (not Wri) then TimeString1 := '';
            Pass := 2;
            Wri := True;
            Reset(InFile);
            Readln(InFile,InString); // Skip first line
            Readln(InFile,InString);
          end; // else
        end; // else
      Le := Length(InString);
      if Le > 23 then if InString[11+TimeStringLe1]='0' then
      begin
        TimeString0 := TimeString1;
        TimeString1 := Copy(InString,1,11+TimeStringLe2);
        if TimeString1 < TimeString0 then case Pass of
          1: begin Wri := True; TimeString0 := ''; end;
          2: InString := '000000000000000000000000'; // End reading, writing
        end; // case
        if Wri then
        begin
          if TimeString1[11] = 'T' then TimeString1[11] := ' ';
          if Interv[Job] = 3600 then TimeString1 := TimeString1 + ':00';
          if Interv[Job] = 86400 then TimeString1 := TimeString1 + '00:00';
          if (TimeString0 <> TimeString1) and (TimeString0 <> '') then
          begin
            OutString := '"' + TimeString0 + '",' + IntToStr(RN) + ',';
            RN := RN + 1;
            for n := 3 to NCols[Job] do
            begin
              OutString := OutString + Col[n];
              if (n < NCols[Job]) then OutString := OutString + ',';
            end;
            Writeln(OutFile,OutString);
            for n := 1 to NCols[Job] do Col[n] := '"NAN"';
          end;
          InData := TakeApart(InString);
          if InData.NFields = 9 then
          begin
            Identif := InData.Fi2;
            if Identif <> '' then
            begin
              for n := 3 to NCols[Job] do
              begin
                if (Identif = ColName[n,Job])
                  and (Col[n] = '"NAN"')
                  and (InData.Fi4 <> 'AL_NUL')
                  and (InData.Fi4 <> 'AL_INVALID')
                  and (InData.Fi4 <> 'AL_OFFLINE')
                  then Col[n] := ChangeToPoint(InData.Fi5);
              end; //for
            end; // if Identif <> ''
          end; // if InData.NFields = 9
        end; // if Wri
      end; // if ... if
    end; // while
    CloseFile(InFile);
    CloseFile(OutFile);
    Form1.Memo1.Lines.Add('Succeded, all files closed');
    Form1.Memo1.Lines.Add('Latest data: '+TimeString0);
    Conver := TimeString0;

  except

    CloseFile(InFile);
    CloseFile(OutFile);
    Form1.Memo1.Lines.Add('Failed, all files closed');
    Conver := '';

  end; // try-except
end; // Conver

function NewInFileName(Name: string): string;
var
  Number,NewName: string;
  L,i,n: Integer;
begin
  Number := '0';
  NewName := Name;
  L := Length(Name);
  if (L > 10) then if (Copy(Name, L-3, 4) <> '.csv') then
  begin
    i := L-1;
    while (i > 9) and (Copy(Name, i, 1) <> '.') do i := i - 1;
    if i > 9 then Number := Copy(Name, i+1, L-i);
  end;
  n := StrToInt(Number);
  if n > 0 then
  begin
    n := n + 1;
    NewName := Copy(Name, 1, i) + IntToStr(n);
    if not FileExists(NewName) then NewName := Copy(Name, 1, i-8) + 'csv';
  end;
  NewInFileName := NewName;
end; // NewInFileName

procedure TForm1.FormCreate(Sender: TObject);
begin
  Init;
end;

procedure TForm1.Timer1Timer(Sender: TObject);
var
  Res1: boolean;
  Res2: string;
  i: integer;
begin
  for i := 1 to NJobs do
  begin
    if (FileAge(InFileName[i])<>PrevAge[i]) and ((Typ[i]=1) or (Typ[i]=2)) then
    begin
      if Typ[i]=2 then Res1 := Analy(i);
      Res2 := Conver(i);
      if Res2 <> '' then
      begin
        PrevAge[i] := FileAge(InFileName[i]);
        StringGrid1.Cells[5,i] := Res2;
        InFileName[i] := NewInFileName(InFileName[i]);
      end;
    end;
  end;
end;

procedure TForm1.StringGrid1SelectCell(Sender: TObject; ACol,
  ARow: Integer; var CanSelect: Boolean);
begin
  {Memo1.Lines.Add('Selected row '+IntToStr(ARow));}
  if Typ[ARow]=3 then Button2.Enabled := True else Button2.Enabled := False;
end;

procedure TForm1.Button1Click(Sender: TObject);
begin
  Form2.Show;
end;

procedure TForm1.Button2Click(Sender: TObject);
var
  Res1: boolean;
  Res2: string;
begin
  Res1 := Analy(StringGrid1.Row);
  Res2 := Conver(StringGrid1.Row);
  if Res2 <> '' then StringGrid1.Cells[5,StringGrid1.Row] := Res2;
end;

end.
